function _inheritsLoose(subClass, superClass) { subClass.prototype = Object.create(superClass.prototype); subClass.prototype.constructor = subClass; _setPrototypeOf(subClass, superClass); }
function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf ? Object.setPrototypeOf.bind() : function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }
import Emitter from './Emitter.js';
/**
 * - Orchestrates data and business logic.
 * - Emits events when data changes.
 * 
 * A `Model` manages an internal table of data attributes and triggers change events when any of its data is modified.  
 * Models may handle syncing data with a persistence layer. To design your models, create atomic, reusable objects 
 * that contain all the necessary functions for manipulating their specific data.  
 * Models should be easily passed throughout your app and used anywhere the corresponding data is needed.  
 * Rasti models stores its attributes in `this.attributes`, which is extended from `this.defaults` and the 
 * constructor `attrs` parameter. For every attribute, a getter is generated to retrieve the model property 
 * from `this.attributes`, and a setter is created to set the model property in `this.attributes` and emit `change` 
 * and `change:attribute` events.
 * @module
 * @extends Rasti.Emitter
 * @param {object} attrs Object containing model attributes to extend `this.attributes`. Getters and setters are generated for `this.attributtes`, in order to emit `change` events.
 * @example
 * import { Model } from 'rasti';
 * // Product model
 * class ProductModel extends Model {
 *     preinitialize() {
 *         // The Product model has `name` and `price` default attributes.
 *         // `defaults` will extend `this.attributes`.
 *         // Getters and setters are generated for `this.attributes`,
 *         // in order to emit `change` events.
 *         this.defaults = {
 *             name: '',
 *             price: 0
 *         };
 *     }
 *
 *     setDiscount(discountPercentage) {
 *         // Apply a discount to the price property.
 *         // This will call a setter that will update `price` in `this.attributes`,
 *         // and emit `change` and `change:price` events.
 *         const discount = this.price * (discountPercentage / 100);
 *         this.price -= discount;
 *     }
 * }
 * // Create a product instance with a name and price.
 * const product = new ProductModel({ name: 'Smartphone', price: 1000 });
 * // Listen to the `change:price` event.
 * product.on('change:price', () => console.log('New Price:', product.price));
 * // Apply a 10% discount to the product.
 * product.setDiscount(10); // Output: "New Price: 900"
 */
var Model = /*#__PURE__*/function (_Emitter) {
  function Model() {
    var _this;
    var attrs = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    _this = _Emitter.call(this) || this;
    // Call preinitialize.
    _this.preinitialize.apply(_this, arguments);
    // attributes object.
    _this.attributes = Object.assign({}, _this.defaults || {}, attrs);
    // Previous attributes.
    _this.previous = {};
    // Generate getters/setters for every attr.
    Object.keys(_this.attributes).forEach(_this.defineAttribute.bind(_this));
    return _this;
  }

  /**
   * If you define a preinitialize method, it will be invoked when the Model is first created, before any instantiation logic is run for the Model.
   * @param {object} attrs Object containing model attributes to extend `this.attributes`.
   */
  _inheritsLoose(Model, _Emitter);
  var _proto = Model.prototype;
  _proto.preinitialize = function preinitialize() {}

  /**
   * Generate getter/setter for the given key. In order to emit `change` events.
   * This method is called internally by the constructor
   * for `this.attributes`.
   * @param {string} key Attribute key.
   */;
  _proto.defineAttribute = function defineAttribute(key) {
    var _this2 = this;
    Object.defineProperty(this, key, {
      get: function get() {
        return _this2.get(key);
      },
      set: function set(value) {
        _this2.set(key, value);
      }
    });
  }

  /**
   * Get an attribute from `this.attributes`.
   * This method is called internally by generated getters.
   * @param {string} key Attribute key.
   * @return {any} The attribute value.
   */;
  _proto.get = function get(key) {
    return this.attributes[key];
  }

  /**
   * Set an attribute into `this.attributes`.  
   * Emit `change` and `change:attribute` if a value change.  
   * Could be called in two forms, `this.set('key', value)` and
   * `this.set({ key : value })`.  
   * This method is called internally by generated setters.  
   * The `change` event listener will receive the model instance, an object containing the changed attributes, and the rest of the arguments passed to `set` method.  
   * The `change:attribute` event listener will receive the model instance, the new attribute value, and the rest of the arguments passed to `set` method.
   * @param {string} key Attribute key or object containing keys/values.
   * @param [value] Attribute value.
   * @return {this} This model.
   * @emits change
   * @emits change:attribute
   */;
  _proto.set = function set(key, value) {
    var _this3 = this;
    var attrs, args;
    // Handle both `"key", value` and `{key: value}` style arguments.
    for (var _len = arguments.length, rest = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
      rest[_key - 2] = arguments[_key];
    }
    if (typeof key === 'object') {
      attrs = key;
      args = [value].concat(rest);
    } else {
      var _attrs;
      attrs = (_attrs = {}, _attrs[key] = value, _attrs);
      args = rest;
    }
    // Are we in a nested `set` call?
    // Calling a `set` inside a `change:attribute` or `change` event listener
    var changing = this._changing;
    this._changing = true;
    // Store changed attributes.
    var changed = {};
    // Store previous attributes.
    if (!changing) {
      this.previous = Object.assign({}, this.attributes);
    }
    // Set attributes.
    Object.keys(attrs).forEach(function (key) {
      // Use equality to determine if value changed.
      if (attrs[key] !== _this3.attributes[key]) {
        changed[key] = attrs[key];
        _this3.attributes[key] = attrs[key];
      }
    });
    var changedKeys = Object.keys(changed);
    // Pending `change` event arguments.
    if (changedKeys.length) this._pending = ['change', this, changed].concat(args);
    // Emit `change:attribute` events.
    changedKeys.forEach(function (key) {
      _this3.emit.apply(_this3, ["change:" + key, _this3, attrs[key]].concat(args));
    });
    // Don't emit `change` event until the end of the nested 
    // `set` calls inside `change:attribute` event listeners.
    if (changing) return this;
    // Emit `change` events, that might be nested.
    while (this._pending) {
      var pendingChange = this._pending;
      this._pending = null;
      this.emit.apply(this, pendingChange);
    }
    // Reset flags.
    this._pending = null;
    this._changing = false;
    return this;
  }

  /**
   * Return object representation of the model to be used for JSON serialization.
   * By default returns `this.attributes`.
   * @return {object} Object representation of the model to be used for JSON serialization.
   */;
  _proto.toJSON = function toJSON() {
    return this.attributes;
  };
  return Model;
}(Emitter);
export { Model as default };