"use strict";

exports.__esModule = true;
exports["default"] = void 0;
/**
 * `Emitter` is a class that provides an easy way to implement the observer pattern 
 * in your applications.  
 * It can be extended to create new classes that have the ability to emit and bind custom named events.   
 * Emitter is used by `Model` and `View` classes, which inherit from it to implement 
 * event-driven functionality.
 *
 * @module
 * @example
 * import { Emitter } from 'rasti';
 * // Custom cart
 * class ShoppingCart extends Emitter {
 *     constructor() {
 *         super();
 *         this.items = [];
 *     }
 *
 *     addItem(item) {
 *         this.items.push(item);
 *         // Emit a custom event called `itemAdded`.
 *         // Pass the added item as an argument to the event listener.
 *         this.emit('itemAdded', item);
 *     }
 * }
 * // Create an instance of ShoppingCart and Logger
 * const cart = new ShoppingCart();
 * // Listen to the `itemAdded` event and log the added item using the logger.
 * cart.on('itemAdded', (item) => {
 *     console.log(`Item added to cart: ${item.name} - Price: $${item.price}`);
 * });
 * // Simulate adding items to the cart
 * const item1 = { name : 'Smartphone', price : 1000 };
 * const item2 = { name : 'Headphones', price : 150 };
 *
 * cart.addItem(item1); // Output: "Item added to cart: Smartphone - Price: $1000"
 * cart.addItem(item2); // Output: "Item added to cart: Headphones - Price: $150"
 */
var Emitter = exports["default"] = /*#__PURE__*/function () {
  function Emitter() {}
  var _proto = Emitter.prototype;
  /**
   * Adds event listener.
   * @param {string} type Type of the event (e.g. `change`).
   * @param {function} listener Callback function to be called when the event is emitted.
   * @example
   * this.model.on('change', this.render.bind(this)); // Re render when model changes.
   */
  _proto.on = function on(type, listener) {
    if (typeof listener !== 'function') {
      throw TypeError('Listener must be a function');
    }
    if (!this.listeners) this.listeners = {};
    if (!this.listeners[type]) this.listeners[type] = [];
    this.listeners[type].push(listener);
  }

  /**
   * Adds event listener that executes once.
   * @param {string} type Type of the event (e.g. `change`).
   * @param {function} listener Callback function to be called when the event is emitted.
   * @example
   * this.model.once('change', () => console.log('This will happen once'));
   */;
  _proto.once = function once(type, _listener2) {
    if (typeof _listener2 === 'function') {
      var self = this;
      var _listener = _listener2;
      _listener2 = function listener() {
        _listener.apply(void 0, arguments);
        self.off(type, _listener2);
      };
    }
    this.on(type, _listener2);
  }

  /**
   * Removes event listeners.
   * @param {string} [type] Type of the event (e.g. `change`). If is not provided, it removes all listeners.
   * @param {function} [listener] Callback function to be called when the event is emitted. If listener is not provided, it removes all listeners for specified type.
   * @example
   * this.model.off('change'); // Stop listening to changes.
   */;
  _proto.off = function off(type, listener) {
    if (!type) {
      this.listeners = {};
    } else {
      if (!listener) {
        delete this.listeners[type];
      } else {
        var listeners = this.listeners[type];
        if (listeners) {
          var copy = listeners.slice();
          copy.forEach(function (fn, idx) {
            if (fn === listener) listeners.splice(idx, 1);
          });
          if (!listeners.length) {
            delete this.listeners[type];
          }
        }
      }
    }
  }

  /**
   * Emits event of specified type. Listeners will receive specified arguments.
   * @param {string} type Type of the event (e.g. `change`).
   * @param {any} [...args] Arguments to be passed to listener.
   * @example
   * this.emit('invalid'); // Emit validation error event.
   */;
  _proto.emit = function emit(type) {
    for (var _len = arguments.length, args = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }
    var listeners = this.listeners && this.listeners[type];
    if (!listeners || !listeners.length) return;
    var copy = listeners.slice();
    copy.forEach(function (fn) {
      fn.apply(void 0, args);
    });
  };
  return Emitter;
}();